/**
 * AutoDAO - Generic DAO on steroids implementation for Java.
 * Copyright (C) 2008  Marat Radchenko <slonopotamus@users.sourceforge.net>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.autodao;

import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import static net.sf.autodao.Utils.getLimitParameterIndex;
import static net.sf.autodao.Utils.getOffsetParameterIndex;
import static net.sf.autodao.Utils.isFinder;
import static net.sf.autodao.Utils.isSingleFind;
import org.aopalliance.intercept.MethodInvocation;
import org.springframework.aop.IntroductionInterceptor;

/** Intercepts finder DAO calls and calls execute*Finder methods on dao. NOT FOR PUBLIC USE. */
final class FinderIntroductionInterceptor implements IntroductionInterceptor {

  /** {@inheritDoc} */
  @Override
  public Object invoke(final MethodInvocation invocation) throws Throwable {
    final FinderExecutor dao = (FinderExecutor) invocation.getThis();
    final Method meth = invocation.getMethod();
    if (isFinder(meth)) {
      final Object[] arguments = invocation.getArguments();
      if (isSingleFind(invocation.getMethod())) {
        return dao.executeSingleFinder(meth, arguments);
      } else {
        //TODO: cache indices between invocations?
        final Integer limitIndex = getLimitParameterIndex(meth);
        final Integer offsetIndex = getOffsetParameterIndex(meth);

        final List<Object> ar = new ArrayList<Object>(Arrays.asList(invocation.getArguments()));
        //This works because we said that limit/offset can only be last two parameters.
        if (limitIndex != null) {
          ar.remove(ar.size() - 1);
        }
        if (offsetIndex != null && !offsetIndex.equals(limitIndex)) {
          ar.remove(ar.size() - 1);
        }

        return dao.executeListFinder(
          meth,
          ar.toArray(new Object[ar.size()]),
          limitIndex == null ? null : (Integer) arguments[limitIndex],
          offsetIndex == null ? null : (Integer) arguments[offsetIndex]
        );
      }
    } else {
      return invocation.proceed();
    }
  }

  /** {@inheritDoc} */
  public boolean implementsInterface(final Class intf) {
    return intf.isInterface() && FinderExecutor.class.isAssignableFrom(intf);
  }
}
