/**
 * AutoDAO - Generic DAO on steroids implementation for Java.
 * Copyright (C) 2008  Marat Radchenko <slonopotamus@users.sourceforge.net>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.autodao.hibernate;

import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;

import org.springframework.beans.factory.support.BeanDefinitionBuilder;
import static org.springframework.beans.factory.support.BeanDefinitionBuilder.genericBeanDefinition;
import org.springframework.beans.factory.xml.AbstractSingleBeanDefinitionParser;
import org.springframework.beans.factory.xml.ParserContext;
import org.springframework.util.StringUtils;
import static org.springframework.util.StringUtils.hasText;
import org.w3c.dom.Element;

/**
 * Processes <code>dao:hibernate</code> elements and builds Spring bean.
 * NOT FOR PUBLIC USE.
 * Creates {@link HibernateDaoImpl} bean and wrap it with {@link HibernateDaoProxyFactoryBean}
 * so that AutoDAO magic works.
 */
public final class HibernateDaoBeanDefinitionParser extends AbstractSingleBeanDefinitionParser {

  /**
   * {@link net.sf.autodao.FinderIntroductionAdvisor} bean id.
   * Stored between invocations so that we create reuse advisor for all DAOs in context.
   */
  private String advisorId;

  /**
   * {@inheritDoc}
   * @return <code>true</code>.
   */
  @Override
  protected boolean shouldGenerateIdAsFallback() {
    return true;
  }

  /** {@inheritDoc} */
  @Override
  protected void doParse(final Element element, final ParserContext context, final BeanDefinitionBuilder builder) {
    final String daoInterfaceName = element.getAttribute("interface");
    if (!StringUtils.hasText(daoInterfaceName)) {
      context.getReaderContext().error("No DAO interface specified", element);
    }
    final Class<?> daoInterface;
    try {
      daoInterface = Class.forName(daoInterfaceName);
    } catch (final ClassNotFoundException e) {
      context.getReaderContext().error("Could not load DAO interface: " + daoInterfaceName, element, e);
      return;
    }

    final String sessionFactoryName = element.getAttribute("session-factory");

    final BeanDefinitionBuilder daoBuilder = genericBeanDefinition("net.sf.autodao.hibernate.HibernateDaoImpl")
      //WARNING! First set sessionFactory, then set allowCreate. Things break otherwise.
      .addPropertyReference("sessionFactory", sessionFactoryName)
      .addPropertyValue("hibernateTemplate.allowCreate", false);

    final String xmlEntityType = element.getAttribute("entity");
    String entityType = null;
    if (hasText(xmlEntityType)) {
      //Explicit entity
      //See ExplicitEntityTest
      entityType = xmlEntityType;
    } else {
      //Automagically discover entity entityType from generics.
      for (final Type t : daoInterface.getGenericInterfaces()) {
        if (t instanceof ParameterizedType) {
          final ParameterizedType pt = (ParameterizedType) t;
          if (pt.getRawType() instanceof Class && ((Class) pt.getRawType()).getName().equals("net.sf.autodao.Dao")) {
            final Type type = pt.getActualTypeArguments()[0];
            if (type instanceof Class) {
              entityType = ((Class) type).getName();
            } else if (type instanceof ParameterizedType) {
              //Special handling for cases like Dao<Entity<Foo>>.
              //See ParamEntityTest
              entityType = ((Class) ((ParameterizedType) type).getRawType()).getName();
            } else {
              //TODO: is it actually possible?
              context.getReaderContext().error(
                "Don't know how to handle " + type + " entity type. Please update to latest AutoDAO release."
                  + " If that doesn't help, please file a bugreport to AutoDAO bugtracker. Thanks.",
                element
              );
            }
            break;
          }
        }
      }
    }
    if (hasText(entityType)) {
      daoBuilder.addPropertyValue("type", entityType);
    } else {
      //See NoEntityTest
      context.getReaderContext().error("No entity type found for " + daoInterfaceName, element);
    }

    if (this.advisorId == null) {
      this.advisorId = context.getReaderContext().registerWithGeneratedName(
        genericBeanDefinition("net.sf.autodao.FinderIntroductionAdvisor").getBeanDefinition()
      );
    }

    builder
      .addPropertyValue("interceptorNames", this.advisorId)
      .addPropertyReference("sessionFactory", sessionFactoryName)
      .addPropertyValue("interfaces", daoInterfaceName)
      .addPropertyValue("type", entityType)
      .addPropertyValue("target", daoBuilder.getBeanDefinition());
  }

  /** {@inheritDoc} */
  @Override
  protected String getBeanClassName(final Element element) {
    return "net.sf.autodao.hibernate.HibernateDaoProxyFactoryBean";
  }
}
