/**
 * AutoDAO - Generic DAO on steroids implementation for Java.
 * Copyright (C) 2008  Marat Radchenko <slonopotamus@users.sourceforge.net>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.autodao.hibernate;

import java.io.Serializable;
import java.lang.reflect.Method;
import java.util.List;

import net.sf.autodao.Dao;
import net.sf.autodao.FinderExecutor;
import net.sf.autodao.PersistentEntity;
import static net.sf.autodao.hibernate.HibernateUtils.getNamedQueryName;
import org.hibernate.Query;
import org.hibernate.Session;
import org.springframework.beans.factory.BeanNameAware;
import org.springframework.beans.factory.annotation.Required;
import org.springframework.orm.hibernate3.HibernateCallback;
import org.springframework.orm.hibernate3.support.HibernateDaoSupport;
import org.springframework.stereotype.Repository;
import org.springframework.util.Assert;

/**
 * Hibernate implementation of AutoDAO. This class is intentionally made final. If you need custom DAO method, introduce
 * additional bean. NOT FOR PUBLIC USE.
 * @param <E>
 * @param <K>
 */
@Repository
public final class HibernateDaoImpl<E extends PersistentEntity<K>, K extends Serializable>
  extends HibernateDaoSupport implements Dao<E, K>, FinderExecutor, BeanNameAware {

  /** class of persistent object. */
  private Class<E> type;

  /** Bean name. */
  private String beanName;

  /** {@inheritDoc} */
  public void create(final E o) {
    this.checkObject(o);
    Assert.isTrue(!this.getHibernateTemplate().contains(o), "You're trying to persist already persistent entity: " + o);
    this.getHibernateTemplate().persist(o);
  }

  /** {@inheritDoc} */
  @SuppressWarnings({ "unchecked" })
  public E get(final K id) {
    Assert.notNull(id, "Id cannot be null");
    return (E) this.getHibernateTemplate().get(this.type, id);
  }

  /** {@inheritDoc} */
  public void update(final E o) {
    this.checkObject(o);
    Assert.isTrue(this.getHibernateTemplate().contains(o), "You're trying to update transient entity: " + o);
    this.getHibernateTemplate().update(o);
  }

  /** {@inheritDoc} */
  public void delete(final E o) {
    this.checkObject(o);
    Assert.isTrue(this.getHibernateTemplate().contains(o), "You're trying to delete transient entity: " + o);
    this.getHibernateTemplate().delete(o);
  }

  /** {@inheritDoc} */
  public void saveOrUpdate(final E o) {
    this.checkObject(o);
    this.getHibernateTemplate().saveOrUpdate(o);
  }

  /** {@inheritDoc} */
  public <T> T executeSingleFinder(final Method method, final Object[] queryArgs) {
    final List<T> result = this.executeListFinder(method, queryArgs, null, null);
    Assert.isTrue(result.size() <= 1, "Non-unique result");
    if (result.isEmpty()) {
      return null;
    } else {
      return result.get(0);
    }
  }

  /** {@inheritDoc} */
  @SuppressWarnings({ "unchecked" })
  public <T> List<T> executeListFinder(final Method method, final Object[] queryArgs,
                                       final Integer limit, final Integer offset) {
    return this.getHibernateTemplate().executeFind(
      new HibernateCallback() {

        public Object doInHibernate(final Session session) {
          final Query q = session.getNamedQuery(
            getNamedQueryName(HibernateDaoImpl.this.getSessionFactory(), HibernateDaoImpl.this.type, method)
          );
          int i = 0;
          for (final Object o : queryArgs) {
            q.setParameter(i, o);
            i++;
          }
          if (limit != null) {
            q.setMaxResults(limit);
          }
          if (offset != null) {
            q.setFirstResult(offset);
          }
          return q.list();
        }
      }
    );
  }

  /**
   * Check entity.
   * @param o entity.
   */
  private void checkObject(final E o) {
    Assert.notNull(o, "Entity cannot be null");
    Assert.isAssignable(this.type, o.getClass(), "Invalid entity type");
  }

  /** {@inheritDoc} */
  @SuppressWarnings({ "unchecked" })
  public E merge(final E o) {
    return (E) this.getHibernateTemplate().merge(o);
  }

  /** {@inheritDoc} */
  public void refresh(final E o) {
    this.getHibernateTemplate().refresh(o);
  }

  /** {@inheritDoc} */
  @Override
  protected void initDao() {
    Assert.notNull(this.type, "You forgot to call #setType() on  bean '" + this.beanName + '\'');
  }

  /** {@inheritDoc} */
  public void setBeanName(final String name) {
    this.beanName = name;
  }

  @Required
  public void setType(final Class<E> type) {
    this.type = type;
  }

  /** {@inheritDoc} */
  @Override
  public String toString() {
    return "AutoDao [" + this.type.getName() + ']';
  }
}
